#!/bin/bash -x

# Note that on a Pi with a 32Gb SD card, with the free space below:

# Filesystem     1K-blocks     Used Available Use% Mounted on
# /dev/mmcblk0p2  29469820 13243184  14709728  48% /

# we actually run out of disk space - not because the disk is full,
# but because we're out of inodes...:

# Filesystem      Inodes   IUsed IFree IUse% Mounted on
# /dev/mmcblk0p2 1839328 1833567  5761  100% /

# Interim solution... zip up all the tile files. (or use my x86 linux,
# but the noise from the fan when running a big job is very annoying)

# Outside requirements: ocrmypdf pdfimages convert deskew zip

if [ "`which convert`" == "" ] ; then
    echo This suite relies heavily on the program \'convert\' from ImageMagick.
    echo You do not appear to have it installed, so please install that first, using:
    echo "   sudo apt-get install imagemagick"
    exit 1
fi

for pfilename in original/IPA-*.jpg; do
    filename="`basename $pfilename .jpg`"
    if [ -f cropped/$filename.png ] ; then
        echo Skipping cropped/$filename.png
    else
        # Remove perforations and unused edges of page scans:
        # 1927x3886
        mkdir -p cropped
        convert original/$filename.jpg -crop 900x4100+1150+0 +repage cropped/$filename.png
    fi

    if [ -f level/$filename.png ] ; then
        echo Skipping level/$filename.png
    else
        if [ "`which deskew`" == "" ] ; then
            echo \# It is essential to ensure that all input images are aligned tightly with
            echo \# the horizontal axis.  We fix up images here using the command \'deskew\'
            echo \# which unfortunately does not have an easy installer.  What you need to
            echo \# do is:
            echo ""
            echo wget https://galfar.vevb.net/wp/download/Deskew-1.30.zip
            echo unzip Deskew-1.30.zip
            UNAME="$(uname -s)"
            if [ "${UNAME}" == "" ] ; then
                UNAME="unknown"
            fi
            case "${UNAME}" in
                Linux*)     machine="Linux";;
                Darwin*)    machine="Mac";;
                CYGWIN*)    machine="Cygwin";;
                MINGW*)     machine="MinGw";;
                MSYS_NT*)   machine="Git";;
                *)          machine="${UNAME}"
            esac
            if [ "$machine" == "Linux" ] ; then
                ARCH="`arch`"
                case $ARCH in
                *"86"*)
                    echo sudo cp Deskew-1.30/Bin/deskew /usr/local/bin/deskew;;
                *"arm"*)
                    echo sudo cp Deskew-1.30/Bin/deskew-arm /usr/local/bin/deskew;;
                *"aarch"*)
                    echo sudo cp Deskew-1.30/Bin/deskew-arm /usr/local/bin/deskew;;
                *)
                    echo \# Depending on the system you are using, issue one of:
                    echo sudo cp Deskew-1.30/Bin/deskew /usr/local/bin/deskew
                    echo sudo cp Deskew-1.30/Bin/deskew-arm /usr/local/bin/deskew
                esac
            elif [ "$machine" == "Mac" ] ; then
                 echo sudo cp Deskew-1.30/Bin/deskew-mac /usr/local/bin/deskew
            else
                echo \# Look in Deskew-1.30/Bin and find the executable that is appropriate
                echo "# for your $machine system, and copy it to your bin directory (often /usr/local/bin)"
                echo \# so that it is executed when you type: deskew
            fi
            echo rm -rf Deskew-1.30 Deskew-1.30.zip
            exit 1
        fi

        # Do small rotations of trimmed page to make it absolutely horizontal:
        mkdir -p level
        deskew -f rgb24 -s s -b FFFFFF -l 0.0001 -o level/$filename.png cropped/$filename.png |& fgrep "[deg]"
    fi


done

#make getlines
cc -Wall -Ofast -o segment segment.c -I/usr/include/netpbm -L/usr/local/netpbm/lib -lnetpbm
#cc -Wall -Wno-unused-but-set-variable -g -o findmean findmean.c -lm
#cc -Wall -o grid2html grid2html.c -lnetpbm
cc -Wall -o reoffset reoffset.c -I/usr/include/netpbm -L/usr/local/netpbm/lib -lnetpbm -lm
if [ -f  prescan.results ] ; then
    echo Prescan was done previously.  Skipping.
else
    cp /dev/null prescan.results
    for pfilename in level/IPA-*.png ; do
        filename="`basename $pfilename .png`"
        convert -threshold 50% $pfilename $filename.pnm
        convert -threshold 50% $pfilename DEBUG.png
        # >&2 echo "$filename"
        # segment.c does part of what getlines used to do -
        # it determines the grid spacing - but it does *not*
        # output any images.
        export LD_LIBRARY_PATH="/usr/local/netpbm/lib"
        ./segment $filename.pnm | tee /dev/stderr >> prescan.results
        rm $filename.pnm
    done
fi
SPACING=( `../../segmenter/findmean < prescan.results` )
echo Using character grid ${SPACING[0]} x ${SPACING[1]}
# We have to re-generate the offsets at the new fixed spacing values,
# then output the gridded page and the lines and the individual chars
mkdir -p grid
mkdir -p tiles
cp /dev/null offset.log
for pfilename in level/IPA-*.png ; do
    filename="`basename $pfilename .png`"
    if [ -f grid/${filename}@*@*.png ] ; then
        echo Skipping grid/${filename}@*@*.png
    else
      convert $pfilename $filename.pnm
      export LD_LIBRARY_PATH="/usr/local/netpbm/lib"
      
      PARAMS=( `./reoffset --xstep ${SPACING[1]} --ystep ${SPACING[0]} $filename.pnm` )
      # IPA-02_1_1.pnm 54.8 6.600000 71 32.7 6.000000 24
      # NOTE: On a demo case where there was only one set of grid spacings, the recalculated
      # offset for that spacing *should* be the same as was used in the only example file,
      # .. so why is it not?
      echo UPDATED PARAMS: ${PARAMS[0]} ${PARAMS[1]} ${PARAMS[2]} ${PARAMS[3]} ${PARAMS[4]} ${PARAMS[5]}
      
      echo $PARAMS >> offset.log
      # getlines should be renamed something like extract,
      # and have the grid calculation part removed.
      # getlines *used to* do the grid determination,
      # but that was moved to segment.c when it was
      # re-written with a better algorithm.
      echo ../../segmenter/getlines --exact-line-spacing=${SPACING[0]} \
           --exact-column-spacing=${SPACING[1]} \
           --line-offset=${PARAMS[2]} \
           --column-offset=${PARAMS[5]} \
           --grid --lines --tiles \
           -v $filename.pnm
      ../../segmenter/getlines --exact-line-spacing=${SPACING[0]} \
           --exact-column-spacing=${SPACING[1]} \
           --line-offset=${PARAMS[2]} \
           --column-offset=${PARAMS[5]} \
           --grid --lines --tiles \
           -v $filename.pnm
      rm $filename.pnm
      find tiles/$filename -name 'col???.pgm' | while read f ; do
        echo convert $f `dirname $f`/`basename $f .pgm`.png \&\& rm $f
        convert $f `dirname $f`/`basename $f .pgm`.png && rm $f
      done
    fi
    LONGNAME="`basename grid/${filename}@*@*.png .png`"
    if [ -f grid/$filename.html ] ; then
        echo Skipping grid/$filename.html
    else
      convert grid/${filename}*.png ${LONGNAME}.pnm
      ../../segmenter/grid2html ${LONGNAME}.pnm > grid/$filename.html
      rm ${LONGNAME}.pnm
    fi
done

   # if [ -f tiles/$filename.zip ] ; then
   #     echo Skipping zip tiles/$filename.zip
   # else
   #     echo "pushd tiles ; zip -r $filename.zip $filename/line* ; popd"
   #     pushd tiles >& /dev/null ; zip -r $filename.zip $filename/line* ; popd >& /dev/null
   #     rm -rf tiles/$filename/line*
   # fi                               

exit 0
